==> beyond-all-reason-RecoilEngine-2025.04.04.tar.gz: RecoilEngine-2025.04.04/rts/Rendering/Map/InfoTexture/Modern/Height.cpp <==
/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#include "Height.h"
#include "Map/HeightLinePalette.h"
#include "Map/ReadMap.h"
#include "Rendering/GlobalRendering.h"
#include "Rendering/Shaders/ShaderHandler.h"
#include "Rendering/Shaders/Shader.h"
#include "System/Color.h"
#include "System/Exceptions.h"
#include "System/Config/ConfigHandler.h"
#include "System/Log/ILog.h"

#include "System/Misc/TracyDefs.h"


// currently defined in HeightLinePalette.cpp
//CONFIG(bool, ColorElev).defaultValue(true).description("If heightmap (default hotkey [F1]) should be colored or not.");



CHeightTexture::CHeightTexture()
: CPboInfoTexture("height")
, CEventClient("[CHeightTexture]", 271990, false)
, needUpdate(true)
{
	eventHandler.AddClient(this);

	texSize = int2(mapDims.mapxp1, mapDims.mapyp1);
	texChannels = 4;

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	RecoilTexStorage2D(GL_TEXTURE_2D, 1, GL_RGBA8, texSize.x, texSize.y);

	glGenTextures(1, &paletteTex);
	glBindTexture(GL_TEXTURE_2D, paletteTex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	RecoilTexStorage2D(GL_TEXTURE_2D, 1, GL_RGBA8, 256, 2);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 1, GL_RGBA, GL_UNSIGNED_BYTE, &CHeightLinePalette::paletteColored[0].r);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 1, 256, 1, GL_RGBA, GL_UNSIGNED_BYTE, &CHeightLinePalette::paletteBlackAndWhite[0].r);

	if (FBO::IsSupported()) {
		fbo.Bind();
		fbo.AttachTexture(texture);
		/*bool status =*/ fbo.CheckStatus("CHeightTexture");
		FBO::Unbind();
	}

	const std::string vertexCode = R"(
		#version 120
		varying vec2 texCoord;

		void main() {
			gl_Position = vec4(0.0, 0.0, 0.0, 1.0);
			gl_Position.xy = gl_Vertex.xy * 2.0 - 1.0;
			texCoord = gl_Vertex.st;
		}
	)";

	const std::string fragmentCode = R"(
		#version 120
		uniform sampler2D texHeight;
		uniform sampler2D texPalette;
		uniform float paletteOffset;
		varying vec2 texCoord;

		void main() {
			float h = texture2D(texHeight, texCoord).r;
			vec2 tc = vec2(h * (8. / 256.), paletteOffset);
			gl_FragColor = texture2D(texPalette, tc);
		}
	)";

	shader = shaderHandler->CreateProgramObject("[CHeightTexture]", "CHeightTexture");
	shader->AttachShaderObject(shaderHandler->CreateShaderObject(vertexCode,   "", GL_VERTEX_SHADER));
	shader->AttachShaderObject(shaderHandler->CreateShaderObject(fragmentCode, "", GL_FRAGMENT_SHADER));
	shader->Link();
	if (!shader->IsValid()) {
		const char* fmt = "%s-shader compilation error: %s";
		LOG_L(L_ERROR, fmt, shader->GetName().c_str(), shader->GetLog().c_str());
	} else {
		shader->Enable();
		shader->SetUniform("texHeight", 0);
		shader->SetUniform("texPalette", 1);
		shader->SetUniform("paletteOffset", configHandler->GetBool("ColorElev") ? 0.0f : 1.0f);
		shader->Disable();
		shader->Validate();
		if (!shader->IsValid()) {
			const char* fmt = "%s-shader validation error: %s";
			LOG_L(L_ERROR, fmt, shader->GetName().c_str(), shader->GetLog().c_str());
		}
	}

	if (!fbo.IsValid() || !shader->IsValid()) {
		throw opengl_error("");
	}
}


void CHeightTexture::UpdateCPU()
{
	RECOIL_DETAILED_TRACY_ZONE;
	const SColor* extraTexPal = CHeightLinePalette::GetData();
	const float* heightMap = readMap->GetCornerHeightMapUnsynced();

	infoTexPBO.Bind();
	infoTexPBO.New(texSize.x * texSize.y * texChannels, GL_STREAM_DRAW);
	auto infoTexMem = reinterpret_cast<SColor*>(infoTexPBO.MapBuffer());

	for (int y = 0; y < texSize.y; ++y) {
		for (int x = 0; x < texSize.x; ++x) {
			const int idx = y * texSize.x + x;
			const float height = heightMap[idx];
			const unsigned int value = ((unsigned int)(height * 8.0f)) % 255;
			infoTexMem[idx] = extraTexPal[value];
		}
	}

	infoTexPBO.UnmapBuffer();
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texSize.x, texSize.y, GL_RGBA, GL_UNSIGNED_BYTE, infoTexPBO.GetPtr());
	infoTexPBO.Invalidate();
	infoTexPBO.Unbind();
}


CHeightTexture::~CHeightTexture()
{
	RECOIL_DETAILED_TRACY_ZONE;
	glDeleteTextures(1, &paletteTex);
	shaderHandler->ReleaseProgramObject("[CHeightTexture]", "CHeightTexture");
}


void CHeightTexture::Update()
{
	RECOIL_DETAILED_TRACY_ZONE;
	needUpdate = false;

	const auto hmTexID = readMap->GetHeightMapTexture();

	if (!fbo.IsValid() || !shader->IsValid() || (hmTexID == 0))
		return UpdateCPU();

	fbo.Bind();
	glViewport(0,0, texSize.x, texSize.y);
	shader->Enable();
	glDisable(GL_BLEND);
	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, paletteTex);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, hmTexID);
	glBegin(GL_QUADS);
		glVertex2f(0.f, 0.f);
		glVertex2f(0.f, 1.f);
		glVertex2f(1.f, 1.f);
		glVertex2f(1.f, 0.f);
	glEnd();
	shader->Disable();
	globalRendering->LoadViewport();
	FBO::Unbind();

	// cleanup
	glActiveTexture(GL_TEXTURE1);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
}


void CHeightTexture::UnsyncedHeightMapUpdate(const SRectangle& rect)
{
	RECOIL_DETAILED_TRACY_ZONE;
	needUpdate = true;
}


bool CHeightTexture::IsUpdateNeeded()
{
	RECOIL_DETAILED_TRACY_ZONE;
	return needUpdate;
}

==> beyond-all-reason-RecoilEngine-2025.04.08.tar.gz: RecoilEngine-2025.04.08/rts/Rendering/Map/InfoTexture/Modern/Height.cpp <==
/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#include "Height.h"
#include "Map/HeightLinePalette.h"
#include "Map/ReadMap.h"
#include "Rendering/GlobalRendering.h"
#include "Rendering/Shaders/ShaderHandler.h"
#include "Rendering/Shaders/Shader.h"
#include "Rendering/GL/SubState.h"
#include "System/Color.h"
#include "System/Exceptions.h"
#include "System/Config/ConfigHandler.h"
#include "System/Log/ILog.h"

#include "System/Misc/TracyDefs.h"


// currently defined in HeightLinePalette.cpp
//CONFIG(bool, ColorElev).defaultValue(true).description("If heightmap (default hotkey [F1]) should be colored or not.");



CHeightTexture::CHeightTexture()
: CModernInfoTexture("height")
, CEventClient("[CHeightTexture]", 271990, false)
, needUpdate(true)
{
	eventHandler.AddClient(this);

	texSize = int2(mapDims.mapxp1, mapDims.mapyp1);

	{
		GL::TextureCreationParams tcp{
			.reqNumLevels = 1,
			.wrapMirror = false,
			.minFilter = GL_NEAREST,
			.magFilter = GL_LINEAR
		};

		texture = GL::Texture2D(texSize, GL_RGBA8, tcp, false);
	}
	{
		GL::TextureCreationParams tcp{
			.reqNumLevels = 1,
			.minFilter = GL_NEAREST,
			.magFilter = GL_LINEAR,
			.wrapModes = std::initializer_list{ GL_REPEAT, GL_CLAMP_TO_EDGE }
		};
		paletteTex = GL::Texture2D(256, 2, GL_RGBA8, tcp, false);
		auto binding = paletteTex.ScopedBind();
		paletteTex.UploadSubImage(CHeightLinePalette::paletteColored      , 0, 0, 256, 1);
		paletteTex.UploadSubImage(CHeightLinePalette::paletteBlackAndWhite, 0, 1, 256, 1);
	}

	CreateFBO("CHeightTexture");

	const std::string fragmentCode = R"(
		#version 130

		uniform sampler2D texHeight;
		uniform sampler2D texPalette;
		uniform float paletteOffset;

		in vec2 uv;
		out vec4 fragData;

		void main() {
			float h = texture(texHeight, uv).r;
			vec2 tc = vec2(h * (8. / 256.), paletteOffset);
			fragData = texture(texPalette, tc);
		}
	)";

	shader = shaderHandler->CreateProgramObject("[CHeightTexture]", "CHeightTexture");
	shader->AttachShaderObject(shaderHandler->CreateShaderObject(vertexCode,   "", GL_VERTEX_SHADER));
	shader->AttachShaderObject(shaderHandler->CreateShaderObject(fragmentCode, "", GL_FRAGMENT_SHADER));
	shader->Link();
	if (!shader->IsValid()) {
		const char* fmt = "%s-shader compilation error: %s";
		LOG_L(L_ERROR, fmt, shader->GetName().c_str(), shader->GetLog().c_str());
	} else {
		shader->Enable();
		shader->SetUniform("texHeight", 0);
		shader->SetUniform("texPalette", 1);
		shader->SetUniform("paletteOffset", configHandler->GetBool("ColorElev") ? 0.0f : 1.0f);
		shader->Disable();
		shader->Validate();
		if (!shader->IsValid()) {
			const char* fmt = "%s-shader validation error: %s";
			LOG_L(L_ERROR, fmt, shader->GetName().c_str(), shader->GetLog().c_str());
		}
	}

	if (!fbo.IsValid() || !shader->IsValid()) {
		throw opengl_error("");
	}
}


void CHeightTexture::UpdateCPU()
{
	RECOIL_DETAILED_TRACY_ZONE;

	static std::vector<SColor> infoTexMem;
	infoTexMem.resize(texSize.x * texSize.y);

	const SColor* extraTexPal = CHeightLinePalette::GetData();
	const float* heightMap = readMap->GetCornerHeightMapUnsynced();

	for (int y = 0; y < texSize.y; ++y) {
		for (int x = 0; x < texSize.x; ++x) {
			const int idx = y * texSize.x + x;
			const float height = heightMap[idx];
			const auto value = static_cast<unsigned int>(height * 8.0f) % 255;
			infoTexMem[idx] = extraTexPal[value];
		}
	}

	auto binding = texture.ScopedBind();
	texture.UploadImage(infoTexMem.data());
}


CHeightTexture::~CHeightTexture()
{
	RECOIL_DETAILED_TRACY_ZONE;
	shaderHandler->ReleaseProgramObject("[CHeightTexture]", "CHeightTexture");
}


void CHeightTexture::Update()
{
	RECOIL_DETAILED_TRACY_ZONE;
	needUpdate = false;

	const auto hmTexID = readMap->GetHeightMapTexture();

	if (!fbo.IsValid() || !shader->IsValid() || (hmTexID == 0))
		return UpdateCPU();

	using namespace GL::State;
	auto state = GL::SubState(
		Blending(GL_FALSE)
	);
	auto binding = paletteTex.ScopedBind(1);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, hmTexID);

	RunFullScreenPass();
}


void CHeightTexture::UnsyncedHeightMapUpdate(const SRectangle& rect)
{
	RECOIL_DETAILED_TRACY_ZONE;
	needUpdate = true;
}


bool CHeightTexture::IsUpdateNeeded()
{
	RECOIL_DETAILED_TRACY_ZONE;
	return needUpdate;
}
